"""
Tracker postprocess tool.
"""
import re
from collections import Counter, defaultdict
from typing import List, Dict, Optional

from tqdm import tqdm

from motrack.config_parser import TrackerPostprocessConfig
from motrack.datasets import BaseDataset
from motrack.evaluation.io import TrackerInferenceWriter, TrackerInferenceReader
from motrack.library.cv.bbox import PredBBox, Point
from motrack.tracker.tracklet import Tracklet

INF = 999_999


def element_distance_from_list(query: int, keys: List[int]) -> int:
    """
    Element distance from the list (keys) is equal to the absolute distance
    to the closest element in the list with the query element

    Args:
        query: Query element
        keys: List of keys

    Returns:
        Distance from the list
    """
    min_distance = None
    for key in keys:
        distance = abs(query - key)
        min_distance = distance if min_distance is None else min(distance, min_distance)

    return min_distance


def find_closest_prev_element(query: int, keys: List[int]) -> int:
    """
    Finds element in the list (keys) that is closest but not greater than the query.

    Args:
        query: Query element
        keys: List of keys

    Returns:
        Closest, not greater element
    """
    value = None
    for key in keys:
        if key > query:
            break
        value = key

    assert value is not None
    return value


def find_closest_next_element(query: int, keys: List[int]) -> int:
    """
    Finds element in the list (keys) that is closest but not less than the query.

    Args:
        query: Query element
        keys: List of keys

    Returns:
        Closest, not less element
    """
    value = None
    for key in keys:
        if key < query:
            continue
        value = key
        break

    assert value is not None
    return value


def interpolate_bbox(start_index: int, start_bbox: PredBBox, end_index: int, end_bbox: PredBBox, index: int) -> PredBBox:
    """
    Perform linear interpolation between two bounding boxes at a specific index.

    Args:
        start_index: The frame index of the starting bounding box.
        start_bbox: The starting bounding box.
        end_index: The frame index of the ending bounding box.
        end_bbox: The ending bounding box.
        index: The target frame index for interpolation.

    Returns:
        Interpolated bounding box at the target index.
    """
    assert start_index < index < end_index, 'The target index must be between start_index and end_index.'

    # Calculate alpha based on the target index
    alpha = (index - start_index) / (end_index - start_index)

    # Interpolate upper left corner
    interpolated_upper_left = Point(
        start_bbox.upper_left.x + alpha * (end_bbox.upper_left.x - start_bbox.upper_left.x),
        start_bbox.upper_left.y + alpha * (end_bbox.upper_left.y - start_bbox.upper_left.y)
    )

    # Interpolate bottom right corner
    interpolated_bottom_right = Point(
        start_bbox.bottom_right.x + alpha * (end_bbox.bottom_right.x - start_bbox.bottom_right.x),
        start_bbox.bottom_right.y + alpha * (end_bbox.bottom_right.y - start_bbox.bottom_right.y)
    )

    # Interpolate label and confidence (if available)
    interpolated_label = start_bbox.label
    interpolated_conf = None
    if start_bbox.conf is not None and end_bbox.conf is not None:
        interpolated_conf = start_bbox.conf + alpha * (end_bbox.conf - start_bbox.conf)

    interpolated_bbox = PredBBox(
        upper_left=interpolated_upper_left,
        bottom_right=interpolated_bottom_right,
        label=interpolated_label,
        conf=interpolated_conf
    )

    return interpolated_bbox


def run_tracker_postprocess(
    dataset: BaseDataset,
    tracker_active_output: str,
    tracker_all_output: str,
    tracker_postprocess_output: str,
    postprocess_cfg: Optional[TrackerPostprocessConfig] = None,
    scene_pattern: str = '(.*?)'
) -> None:
    """
    Performs postprocess on tracker inference output files.

    Args:
        dataset: Dataset
        tracker_active_output: Path where tracker ACTIVE tracks inference can be found
        tracker_all_output: Path where tracker ALL tracks inference can be found (NEW, ACTIVE, LOST)
        tracker_postprocess_output: Path where the postprocessed files will be stored
        postprocess_cfg: Custom postprocess config (if not set then default is used)
        scene_pattern: Scene pattern filter (regex)
    """
    if postprocess_cfg is None:
        postprocess_cfg = TrackerPostprocessConfig()

    scene_names = dataset.scenes
    scene_names = [scene_name for scene_name in scene_names if re.match(scene_pattern, scene_name)]
    for scene_name in tqdm(scene_names, desc='Postprocessing tracker', unit='scene'):
        scene_info = dataset.get_scene_info(scene_name)
        scene_length = scene_info.seqlength
        imheight = scene_info.imheight
        imwidth = scene_info.imwidth

        tracklet_presence_counter = Counter()  # Used to visualize new, appearing tracklets
        tracklet_frame_bboxes: Dict[str, Dict[int, PredBBox]] = defaultdict(dict)
        with TrackerInferenceReader(tracker_active_output, scene_name, image_height=imheight, image_width=imwidth) as tracker_inf_reader:
            last_read = tracker_inf_reader.read()

            for index in tqdm(range(scene_length), desc=f'Counting occurrences "{scene_name}"', unit='frame'):
                if last_read is not None and index == last_read.frame_index:
                    for tracklet_id, bbox in last_read.objects.items():
                        tracklet_presence_counter[tracklet_id] += 1
                        tracklet_frame_bboxes[tracklet_id][index] = bbox

                    last_read = tracker_inf_reader.read()

        # (1) Filter short tracklets (they have less than X active frames)
        tracklets_to_keep = {k for k, v in tracklet_presence_counter.items() if v >= postprocess_cfg.min_tracklet_length}
        tracklet_frame_bboxes = dict(tracklet_frame_bboxes)

        with (TrackerInferenceReader(tracker_all_output, scene_name, image_height=imheight, image_width=imwidth) as tracker_all_inf_reader, \
            TrackerInferenceWriter(tracker_postprocess_output, scene_name, image_height=imheight, image_width=imwidth) as tracker_inf_writer):

            last_all_read = tracker_all_inf_reader.read()

            for index in tqdm(range(scene_length), desc=f'Postprocessing "{scene_name}"', unit='frame'):
                if last_all_read is not None and index == last_all_read.frame_index:
                    for tracklet_id, bbox in last_all_read.objects.items():

                        if tracklet_id not in tracklets_to_keep:
                            # Marked as FP in postprocessing
                            continue

                        tracklet_indices = list(tracklet_frame_bboxes[tracklet_id].keys())
                        keep = False
                        if index in tracklet_indices:
                            keep = True
                        else:
                            # (2) Linear interpolation
                            if min(tracklet_indices) <= index <= max(tracklet_indices) \
                                and tracklet_presence_counter[tracklet_id] >= postprocess_cfg.linear_interpolation_min_tracklet_length:
                                prev_index = find_closest_prev_element(index, tracklet_indices)
                                next_index = find_closest_next_element(index, tracklet_indices)
                                if next_index - prev_index <= postprocess_cfg.linear_interpolation_threshold:
                                    bbox = interpolate_bbox(
                                        start_index=prev_index,
                                        start_bbox=tracklet_frame_bboxes[tracklet_id][prev_index],
                                        end_index=next_index,
                                        end_bbox=tracklet_frame_bboxes[tracklet_id][next_index],
                                        index=index
                                    )
                                    keep = True

                        # (3) Add trajectory initialization
                        start_index = min(tracklet_indices)
                        if index < start_index and start_index - index <= postprocess_cfg.init_threshold:
                            keep = True

                        if keep:
                            tracklet = Tracklet(bbox=bbox, frame_index=index, _id=int(tracklet_id))
                            tracker_inf_writer.write(frame_index=index, tracklet=tracklet)

                    last_all_read = tracker_all_inf_reader.read()



"""
Object detection inference interface.
"""
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Any

import numpy as np

from motrack.library.cv.bbox import PredBBox, BBox
from motrack.utils.lookup import LookupTable


class ObjectDetectionInference(ABC):
    """
    ObjectDetection inference interface.
    """
    def __init__(self, lookup: Optional[LookupTable] = None):
        self._lookup = lookup

    @abstractmethod
    def predict_raw(self, image: np.ndarray) -> Any:
        """
        Performs inference and returns raw model output.

        Args:
            image: Image data

        Returns:
            Raw model output
        """

    @abstractmethod
    def postprocess(self, image: np.ndarray, raw: Any) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Performs postprocess of the raw model output.

        Args:
            image: Image data
            raw: Raw model output

        Returns:
            - List of object bboxes arrays in format xyxy
            - List of object classes
            - List of object bbox confidences
        """

    def predict_with_postprocess(self, image: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Generates list of raw bbox prediction based on given image data.

        Args:
            image: Image data

        Returns:
            - List of object bboxes arrays in format xyxy
            - List of object classes (ints)
            - List of object bbox confidences
        """
        raw = self.predict_raw(image)
        return self.postprocess(image, raw)

    def predict_bboxes(self, image: np.ndarray) -> List[PredBBox]:
        """
        Generates list of bbox prediction based on given image data

        Args:
            image: Image data

        Returns:
            List of bbox predictions
        """
        bboxes, classes, confidences = self.predict_with_postprocess(image)
        return self.pack_bboxes(bboxes, classes, confidences)

    def pack_bboxes(
        self,
        bbox_xyxy: np.ndarray,
        classes: np.ndarray,
        confidences: np.ndarray
    ) -> List[PredBBox]:
        """
        Packs OD raw data into bboxes

        Args:
            bbox_xyxy: BBox in xyxy coordinate format
            classes: Classes (int or str)
            confidences: Confidences

        Returns:
            Returns list of bboxes
        """
        classes = [int(cls) for cls in classes]
        confidences = [float(conf) for conf in confidences]
        if self._lookup is not None:
            classes = [self._lookup.inverse_lookup(cls) for cls in classes]

        pred_bboxes: List[PredBBox] = []
        for bbox_data, cls, conf in zip(bbox_xyxy, classes, confidences):
            pred_bbox = PredBBox.create(
                bbox=BBox.from_xyxy(*bbox_data),
                label=cls,
                conf=conf
            )
            pred_bboxes.append(pred_bbox)

        return pred_bboxes



"""
Dataset interface (base class)
"""
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional, Union, Tuple

import cv2
import numpy as np


@dataclass
class BasicSceneInfo:
    """
    Basic scene information
    """
    name: str
    category: str
    seqlength: Union[int, str]  # Can be parsed from `str`
    imheight: Union[int, str]
    imwidth: Union[int, str]

    def __post_init__(self):
        """
        Convert to proper type.
        """
        self.seqlength = int(self.seqlength)
        self.imheight = int(self.imheight)
        self.imwidth = int(self.imwidth)


@dataclass
class ObjectFrameData:
    frame_id: int
    bbox: Union[List[float], np.ndarray]
    image_path: str
    scene: Union[str, int]
    category: Union[str, int]
    occ: bool
    oov: bool


class BaseDataset(ABC):
    """
    Defines dataset interface.
    """
    def __init__(
        self,
        test: bool = False,
        sequence_list: Optional[List[str]] = None,
        image_shape: Union[None, List[int], Tuple[int, int]] = None,
        image_bgr_to_rgb: bool = True
    ):
        self._test = test
        self._sequence_list = sequence_list

        if image_shape is not None:
            assert isinstance(image_shape, (list, tuple)), \
                f'Invalid image shape type "{type(image_shape)}".'
            image_shape = tuple(image_shape)
            assert len(image_shape) == 2, f'Invalid image shape length "{len(image_shape)}"'

        self._image_shape = image_shape
        self._image_bgr_to_rgb = image_bgr_to_rgb

    def load_image(self, path: str) -> np.ndarray:
        """
        Loads image using cv2.

        Args:
            path: Image source path

        Returns:
            Loaded image
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f'Failed to load image! File not found "{path}".')
        image = cv2.imread(path)
        assert image is not None, 'Invalid Program State!'

        if self._image_bgr_to_rgb:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        if self._image_shape is not None:
            image = cv2.resize(image, self._image_shape, interpolation=cv2.INTER_NEAREST)
        return image

    @abstractmethod
    def load_scene_image_by_frame_index(self, scene_name: str, frame_index: int) -> np.ndarray:
        """
        Loads scene image chosen by frame

        Args:
            scene_name: Scene name
            frame_index: Frame index

        Returns:
            Loaded image
        """

    @abstractmethod
    def __len__(self) -> int:
        pass

    @property
    @abstractmethod
    def scenes(self) -> List[str]:
        """
        Returns:
            List of scenes in dataset.
        """

    @abstractmethod
    def parse_object_id(self, object_id: str) -> Tuple[str, str]:
        """
        Parses and validates object id.

        Object id convention is `{scene_name}_{scene_object_id}` and is unique over all scenes.

        For MOT {scene_name} represents one video sequence.
        For SOT {scene_name} does not need to be unique for the sequence
        but `{scene_name}_{scene_object_id}` is always unique

        Args:
            object_id: Object id

        Returns:
            scene name, scene object id
        """

    @abstractmethod
    def get_object_category(self, object_id: str) -> str:
        """
        Gets category for object.

        Args:
            object_id: Object id

        Returns:
            Object category
        """

    @abstractmethod
    def get_scene_object_ids(self, scene_name: str) -> List[str]:
        """
        Gets object ids for given scene name

        Args:
            scene_name: Scene name

        Returns:
            Scene objects
        """

    def get_scene_number_of_object_ids(self, scene_name: str) -> int:
        """
        Gets number of unique objects in the scene.

        Args:
            scene_name: Scene name

        Returns:
            Number of objects in the scene
        """
        return len(self.get_scene_object_ids(scene_name))

    @abstractmethod
    def get_object_data_length(self, object_id: str) -> int:
        """
        Gets total number of data points for given `object_id` for .

        Args:
            object_id: Object id

        Returns:
            Number of data points
        """

    @abstractmethod
    def get_object_data(self, object_id: str, index: int, relative_bbox_coords: bool = True) -> ObjectFrameData:
        """
        Get object data point index.

        Args:
            object_id: Object id
            index: Index
            relative_bbox_coords: Scale bbox coords to [0, 1]

        Returns:
            Data point.
        """

    @abstractmethod
    def get_object_data_by_frame_index(
        self,
        object_id: str,
        frame_index: int,
        relative_bbox_coords: bool = True
    ) -> Optional[ObjectFrameData]:
        """
        Like `get_object_data_label` but data is relative to given frame_index.
        If object does not exist in given frame index then None is returned.

        Args:
            object_id: Object id
            frame_index: Frame Index
            relative_bbox_coords: Scale bbox coords to [0, 1]

        Returns:
            Data point.
        """

    @abstractmethod
    def get_scene_info(self, scene_name: str) -> BasicSceneInfo:
        """
        Get scene metadata by name.

        Args:
            scene_name: Scene name

        Returns:
            Scene metadata
        """

    @abstractmethod
    def get_scene_image_path(self, scene_name: str, frame_index: int) -> str:
        """
        Get image (frame) path for given scene and frame id.

        Args:
            scene_name: scene name
            frame_index: frame index

        Returns:
            Frame path
        """



"""
Pipeline (tools scripts) functions.
"""
import logging
import os
import time
from datetime import datetime
from pathlib import Path
from typing import Callable

from omegaconf import DictConfig, OmegaConf

from motrack.common import project, formats
from motrack.utils import rich

logger = logging.getLogger('PipelineUtils')


def task(task_name: str) -> Callable:
    """
    Optional decorator that wraps the task function in extra utilities.

    Args:
        task_name: Task name

    Returns:
        Task wrapper decorator
    """
    def task_wrapper(task_func: Callable) -> Callable:
        """
        Args:
            task_func: Function to wrap

        Returns:
            Wrapped function
        """

        def wrap(cfg: DictConfig):
            # Extracting `master_path` from optional `cfg.path.master`.
            master_path = None
            paths = cfg.get('path')
            if paths is not None:
                master_path = paths.get('master')
            master_path = project.OUTPUTS_PATH if master_path is None else master_path
            Path(master_path).mkdir(parents=True, exist_ok=True)

            # Print config
            rich.print_config_tree(cfg, resolve=True, save_to_file=True)

            # Store config history
            store_run_history_config(master_path, cfg, task_name=task_name)

            # execute the task
            start_time = time.time()
            # Parse config
            cfg = OmegaConf.to_object(cfg)

            # Run
            task_func(cfg=cfg)
            logger.info(f"'{task_func.__name__}' execution time: {time.time() - start_time} (s)")

        return wrap

    return task_wrapper


def store_run_history_config(output_dir: str, cfg: DictConfig, task_name: str) -> None:
    """
    Stores run config of the task run.

    Args:
        output_dir: Task output path
        cfg: Task config
        task_name: Task name
    """
    config_dirpath = os.path.join(output_dir, cfg.dataset.type, cfg.algorithm.name, cfg.experiment, 'run_configs')
    dt = datetime.now().strftime(formats.DATETIME_FORMAT)
    config_path = os.path.join(config_dirpath, f'{dt}_{task_name}.yaml')
    Path(config_dirpath).mkdir(parents=True, exist_ok=True)
    with open(config_path, 'w', encoding='utf-8') as f:
        f.write(OmegaConf.to_yaml(cfg))



"""
Numpy IO functions.
"""
from pathlib import Path

import numpy as np


def load_npy(path: str) -> np.ndarray:
    """
    Loads numpy array from given path.

    Args:
        path: Path

    Returns:
        Loaded numpy array.
    """
    with open(path, 'rb') as f:
        return np.load(f)


def store_npy(data: np.ndarray, path: str) -> None:
    """
    Stores numpy array in path.

    Args:
        data: Data
        path: Path
    """
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    with open(path, 'wb') as f:
        np.save(f, data)



"""
Defines interface for motion model filter (predict - update)
"""
from abc import ABC, abstractmethod
from typing import Tuple, Any

import numpy as np

State = Any


class StateModelFilter(ABC):
    """
    Defines interface for motion model filter (predict - update)
    """
    @abstractmethod
    def initiate(self, measurement: np.ndarray) -> State:
        """
        Initializes model state with initial measurement.

        Args:
            measurement: Starting measurement

        Returns:
            Initial state
        """
        pass

    @abstractmethod
    def predict(self, state: State) -> State:
        """
        Predicts prior state.

        Args:
            state: Previous posterior state.

        Returns:
            Prior state (prediction)
        """
        pass

    @abstractmethod
    def multistep_predict(self, state: State, n_steps: int) -> State:
        """
        Estimates prior state for multiple steps ahead

        Args:
            state: Current state
            n_steps: Number of prediction steps

        Returns:
            Estimations for n_steps ahead
        """
        pass

    @abstractmethod
    def update(self, state: State, measurement: np.ndarray) -> State:
        """
        Updates state model based on new measurement.

        Args:
            state: Prior state.
            measurement: New measurement

        Returns:
            Posterior state.
        """
        pass

    @abstractmethod
    def missing(self, state: State) -> State:
        """
        Update state when measurement is missing (replacement for update function for that case).

        Args:
            state: State (prior)

        Returns:
            Updated state
        """

    @abstractmethod
    def project(self, state: State) -> Tuple[np.ndarray, np.ndarray]:
        """
        Projects state to the measurement space (estimation with uncertainty)

        Args:
            state: Model state

        Returns:
            Mean, Covariance
        """
        pass

    @abstractmethod
    def singlestep_to_multistep_state(self, state: State) -> State:
        """
        Converts singlestep state to multistep state. Used as an optimization.

        Args:
            state: Single step state

        Returns:
            Multistep state
        """
        pass

    def affine_transform(self, state: State, warp: np.ndarray) -> State:
        """
        Applies affine transform to the filter state.
        This is required only when used with CMC.

        Args:
            state: Current state
            warp: Affine transform matrix (warp)

        Returns:
            Warped state
        """
        raise NotImplementedError('This filter does not support affine transform!')



"""
Wrapper of Bot-Sort KF.
"""
from typing import Tuple, Optional

import numpy as np

from motrack.filter.algorithms.base import State
from motrack.filter.algorithms.base import StateModelFilter
from motrack.filter.catalog import FILTER_CATALOG
from motrack.library.kalman_filter.botsort_kf import BotSortKalmanFilter


@FILTER_CATALOG.register('bot-sort')
class BotSortKalmanWrapFilter(StateModelFilter):
    """
    Wrapper for BotSortKalman filter for StateModelFilter interface.
    """
    def __init__(
        self,
        override_std_weight_position: Optional[float] = None
    ):
        self._kf = BotSortKalmanFilter(
            override_std_weight_position=override_std_weight_position
        )

    def initiate(self, measurement: np.ndarray) -> State:
        mean, covariance = self._kf.initiate(measurement)
        return mean, covariance

    def predict(self, state: State) -> State:
        mean, covariance = state
        mean_hat, covariance_hat = self._kf.predict(mean, covariance)
        return mean_hat, covariance_hat

    def multistep_predict(self, state: State, n_steps: int) -> State:
        mean, covariance = state
        mean_hat_list, cov_hat_list = [], []
        for _ in range(n_steps):
            mean, covariance = self.predict((mean, covariance))
            mean_hat_list.append(mean)
            cov_hat_list.append(covariance)

        mean_hat = np.stack(mean_hat_list)
        covariance_hat = np.stack(cov_hat_list)
        return mean_hat, covariance_hat

    def update(self, state: State, measurement: np.ndarray) -> State:
        mean_hat, covariance_hat = state
        mean, covariance = self._kf.update(mean_hat, covariance_hat, measurement)
        return mean, covariance

    def singlestep_to_multistep_state(self, state: State) -> State:
        mean, covariance = state
        return mean.unsqueeze(0), covariance.unsqueeze(0)

    def missing(self, state: State) -> State:
        return state  # Use prior instead of posterior

    def project(self, state: State) -> Tuple[np.ndarray, np.ndarray]:
        mean, covariance = state
        if len(mean.shape) == 1:
            # Single step projection
            mean_proj, covariance_proj = self._kf.project(mean, covariance)
        elif len(mean.shape) == 2:
            # Multistep (batch) projection
            batch_size = mean.shape[0]
            mean_proj_list, cov_proj_list = [], []
            for i in range(batch_size):
                mean_i, covariance_i = mean[i, :], covariance[i, :, :]
                mean_proj_i, covariance_proj_i = self._kf.project(mean_i, covariance_i)
                mean_proj_list.append(mean_proj_i)
                cov_proj_list.append(covariance_proj_i)

            mean_proj = np.stack(mean_proj_list)
            covariance_proj = np.stack(cov_proj_list)
        else:
            raise AssertionError(f'Invalid shape {mean.shape}')

        return mean_proj, np.diagonal(covariance_proj, axis1=-2, axis2=-1)

    def affine_transform(self, state: State, warp: np.ndarray) -> State:
        measurement, covariance = state
        L = np.kron(np.eye(4, dtype=np.float32), warp[:, :2])
        T = np.zeros(shape=8, dtype=np.float32)
        T[0:2] = warp[:, 2]
        measurement = L @ measurement + T
        covariance = L @ covariance @ L.T
        return measurement, covariance


def run_test() -> None:
    # TODO: Move to unit tests
    smf = BotSortKalmanWrapFilter()
    measurements = np.random.randn(10, 4)

    # Initiate test
    mean, cov = smf.initiate(measurements[0])
    assert mean.shape == (8,) and cov.shape == (8, 8)

    # Predict test
    mean_hat, cov_hat = smf.predict((mean, cov))
    assert mean_hat.shape == (8,) and cov_hat.shape == (8, 8)

    # Projected predict
    mean_hat_proj, cov_hat_proj = smf.project((mean_hat, cov_hat))
    assert mean_hat_proj.shape == (4,) and cov_hat_proj.shape == (4, 4)

    # Update test
    mean_updated, cov_updated = smf.update((mean, cov), measurements[1])
    assert mean_updated.shape == (8,) and cov_updated.shape == (8, 8)

    # Projected update
    mean_updated_proj, cov_updated_proj = smf.project((mean_updated, cov_updated))
    assert mean_updated_proj.shape == (4,) and cov_updated_proj.shape == (4, 4)

    # Multistep predict
    mean_multistep_hat, cov_multistep_hat = smf.multistep_predict((mean, cov), n_steps=5)
    assert mean_multistep_hat.shape == (5, 8) and cov_multistep_hat.shape == (5, 8, 8)

    # Projected multistep
    mean_multistep_hat_proj, cov_multistep_hat_proj = smf.project((mean_multistep_hat, cov_multistep_hat))
    assert mean_multistep_hat_proj.shape == (5, 4) and cov_multistep_hat_proj.shape == (5, 4, 4)


if __name__ == '__main__':
    run_test()



"""
Tracker. Supports:
- SORT
- ByteTrack
- SparseTrack
"""
from motrack.tracker.matching import AssociationAlgorithm, IoUAssociation
from motrack.tracker.trackers.algorithms.base import Tracker
from motrack.tracker.trackers.algorithms.byte import ByteTracker
from motrack.tracker.trackers.algorithms.sort import SortTracker
from motrack.tracker.trackers.algorithms.sparse import SparseTracker
from motrack.tracker.trackers.factory import tracker_factory



"""
MOT Challenge Dataset support. Supports: MOT17, MOT20, DanceTrack.
"""
import configparser
import copy
import logging
import os
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Tuple, List, Union, Optional

import numpy as np
import pandas as pd
from tqdm import tqdm

from motrack.datasets.base import BaseDataset, BasicSceneInfo, ObjectFrameData
from motrack.datasets.catalog import DATASET_CATALOG
from motrack.utils import file_system

CATEGORY = 'pedestrian'
N_IMG_DIGITS = 6


@dataclass
class SceneInfo(BasicSceneInfo):
    """
    MOT Scene metadata (name, frame shape, ...)
    """
    dirpath: str
    gt_path: str
    framerate: Union[str, int]
    imdir: str
    imext: str

    def __post_init__(self):
        """
        Convert to proper type
        """
        super().__post_init__()
        self.framerate = int(self.framerate)


SceneInfoIndex = Dict[str, SceneInfo]


logger = logging.getLogger('MOTDataset')


@DATASET_CATALOG.register('mot')
class MOTDataset(BaseDataset):
    """
    Parses MOT dataset in given format
    """
    def __init__(
        self,
        path: str,
        sequence_list: Optional[List[str]] = None,
        image_shape: Union[None, List[int], Tuple[int, int]] = None,
        image_bgr_to_rgb: bool = True,
        test: bool = False
    ) -> None:
        """
        Args:
            path: Path to dataset
            sequence_list: Sequence filter by defined list
            image_shape: Resize images (optional - default: no resize)
            image_bgr_to_rgb: Convert BGR to RGB
        """
        super().__init__(
            test=test,
            sequence_list=sequence_list,
            image_shape=image_shape,
            image_bgr_to_rgb=image_bgr_to_rgb
        )

        self._path = path

        self._scene_info_index, self._n_digits = self._index_dataset(path, sequence_list, test=test)
        self._data_labels, self._n_labels, self._frame_to_data_index_lookup = self._parse_labels(self._scene_info_index, test=test)

    @property
    def scenes(self) -> List[str]:
        return list(self._scene_info_index.keys())

    def parse_object_id(self, object_id: str) -> Tuple[str, str]:
        assert object_id in self._data_labels, f'Unknown object id "{object_id}".'
        scene_name, scene_object_id = object_id.split('_')
        return scene_name, scene_object_id

    def get_object_category(self, object_id: str) -> str:
        return CATEGORY

    def get_scene_object_ids(self, scene_name: str) -> List[str]:
        assert scene_name in self.scenes, f'Unknown scene "{scene_name}". Dataset scenes: {self.scenes}.'
        return [d for d in self._data_labels if d.startswith(scene_name)]

    def get_scene_number_of_object_ids(self, scene_name: str) -> int:
        return len(self.get_scene_object_ids(scene_name))

    def get_object_data_length(self, object_id: str) -> int:
        return len(self._data_labels[object_id])

    def get_object_data(self, object_id: str, index: int, relative_bbox_coords: bool = True) -> Optional[ObjectFrameData]:
        data = copy.deepcopy(self._data_labels[object_id][index])
        if data is None:
            return None

        if relative_bbox_coords:
            scene_name, _ = self.parse_object_id(object_id)
            scene_info = self._scene_info_index[scene_name]
            data.bbox[0] = data.bbox[0] / scene_info.imwidth
            data.bbox[1] = data.bbox[1] / scene_info.imheight
            data.bbox[2] = data.bbox[2] / scene_info.imwidth
            data.bbox[3] = data.bbox[3] / scene_info.imheight

        return data

    def load_scene_image_by_frame_index(self, scene_name: str, frame_index: int) -> np.ndarray:
        # noinspection PyTypeChecker
        scene_info: SceneInfo = self.get_scene_info(scene_name)
        image_path = self._get_image_path(scene_info, frame_index + 1)
        return self.load_image(image_path)

    def get_object_data_by_frame_index(
        self,
        object_id: str,
        frame_index: int,
        relative_bbox_coords: bool = True
    ) -> Optional[dict]:
        return self.get_object_data(object_id, frame_index, relative_bbox_coords=relative_bbox_coords)

    def get_scene_info(self, scene_name: str) -> BasicSceneInfo:
        return self._scene_info_index[scene_name]

    def _get_image_path(self, scene_info: SceneInfo, frame_id: int) -> str:
        """
        Get frame path for given scene and frame id

        Args:
            scene_info: scene metadata
            frame_id: frame number

        Returns:
            Path to image (frame)
        """
        return os.path.join(scene_info.dirpath, scene_info.imdir, f'{frame_id:0{self._n_digits}d}{scene_info.imext}')

    def get_scene_image_path(self, scene_name: str, frame_index: int) -> str:
        scene_info = self._scene_info_index[scene_name]
        return self._get_image_path(scene_info, frame_index + 1)

    @staticmethod
    def _get_data_cache_path(path: str, data_name: str) -> str:
        """
        Get cache path for data object path.

        Args:
            path: Path

        Returns:
            Path where data object is or will be stored.
        """
        filename = Path(path).stem
        cache_filename = f'.{filename}.{data_name}.json'
        dirpath = str(Path(path).parent)
        return os.path.join(dirpath, cache_filename)

    @staticmethod
    def parse_scene_ini_file(scene_directory: str, label_type_name) -> SceneInfo:
        gt_path = os.path.join(scene_directory, label_type_name, f'{label_type_name}.txt')
        scene_info_path = os.path.join(scene_directory, 'seqinfo.ini')
        raw_info = configparser.ConfigParser()
        raw_info.read(scene_info_path)
        raw_info = dict(raw_info['Sequence'])
        raw_info['gt_path'] = gt_path
        raw_info['dirpath'] = scene_directory

        return SceneInfo(**raw_info, category=CATEGORY)

    @staticmethod
    def _index_dataset(
        path: str,
        sequence_list: Optional[List[str]],
        test: bool = False
    ) -> Tuple[SceneInfoIndex, int]:
        """
        Index dataset content. Format: { {scene_name}: {scene_labels_path} }

        Args:
            path: Path to dataset
            sequence_list: Filter scenes
            test: Is it test split

        Returns:
            Index to scenes, number of digits used in images name convention (may vary between datasets)
        """
        scene_names = [file for file in file_system.listdir(path) if not file.startswith('.')]
        logger.debug(f'Found {len(scene_names)} scenes. Names: {scene_names}.')
        n_digits = N_IMG_DIGITS

        scene_info_index: SceneInfoIndex = {}
        n_skipped: int = 0

        for scene_name in scene_names:
            if sequence_list is not None and scene_name not in sequence_list:
                continue

            scene_directory = os.path.join(path, scene_name)
            scene_files = file_system.listdir(scene_directory)

            # Scene content validation
            skip_scene = False
            files_to_check = ['gt', 'seqinfo.ini'] if not test else ['seqinfo.ini']
            for filename in files_to_check:
                if filename not in scene_files:
                    msg = f'Ground truth file "{filename}" not found on path "{scene_directory}". Contents: {scene_files}'
                    raise FileNotFoundError(msg)

            if 'img1' in scene_files:
                # Check number of digits used in image name (e.g. DanceTrack and MOT20 have different convention)
                img1_path = os.path.join(scene_directory, 'img1')
                image_names = file_system.listdir(img1_path)
                assert len(image_names) > 0, 'Image folder exists but it is empty!'
                image_name = Path(image_names[0]).stem
                n_digits = len(image_name)

            if skip_scene:
                n_skipped += 1
                continue

            scene_info = MOTDataset.parse_scene_ini_file(scene_directory, 'gt')
            scene_info_index[scene_name] = scene_info
            logger.debug(f'Scene info {scene_info}.')

        if n_digits != N_IMG_DIGITS:
            logger.warning(f'This dataset does not have default number of digits in image name. Got {n_digits} where default is {N_IMG_DIGITS}.')

        logger.info(f'Total number of parsed scenes is {len(scene_info_index)}. Number of skipped scenes is {n_skipped}.')

        return scene_info_index, n_digits

    def _parse_labels(self, scene_infos: SceneInfoIndex, test: bool = False) -> Tuple[Dict[str, List[ObjectFrameData]], int, Dict[str, Dict[int, int]]]:
        """
        Loads all labels dictionary with format:
        {
            {scene_name}_{object_id}: {
                {frame_id}: [ymin, xmin, w, h]
            }
        }

        Args:
            scene_infos: Scene Metadata
            test: If test then no parsing is performed

        Returns:
            Labels dictionary
        """
        data: Dict[str, List[Optional[ObjectFrameData]]] = {}
        frame_to_data_index_lookup: Dict[str, Dict[int, int]] = defaultdict(dict)
        n_labels = 0
        if test:
            # Return empty labels
            return data, n_labels, frame_to_data_index_lookup

        for scene_name, scene_info in scene_infos.items():
            seqlength = self._scene_info_index[scene_name].seqlength

            df = pd.read_csv(scene_info.gt_path, header=None)
            df = df[df[7] == 1]  # Ignoring values that are not evaluated

            df = df.iloc[:, :6]
            df.columns = ['frame_id', 'object_id', 'xmin', 'ymin', 'w', 'h']  # format: yxwh
            df['object_global_id'] = \
                scene_name + '_' + df['object_id'].astype(str)  # object id is not unique over all scenes
            df = df.drop(columns='object_id', axis=1)
            df = df.sort_values(by=['object_global_id', 'frame_id'])
            n_labels += df.shape[0]

            object_groups = df.groupby('object_global_id')
            for object_global_id, df_grp in tqdm(object_groups, desc=f'Parsing {scene_name}', unit='pedestrian'):
                df_grp = df_grp.drop(columns='object_global_id', axis=1).set_index('frame_id')

                data[object_global_id] = [None for _ in range(seqlength)]
                for frame_id, row in df_grp.iterrows():
                    data[object_global_id][int(frame_id) - 1] = ObjectFrameData(
                        frame_id=frame_id,
                        bbox=row.values.tolist(),
                        image_path=self._get_image_path(scene_info, frame_id),
                        occ=False,
                        oov=False,
                        scene=scene_name,
                        category=CATEGORY
                    )
                    frame_to_data_index_lookup[object_global_id][frame_id] = len(data[object_global_id]) - 1

        logger.debug(f'Parsed labels. Dataset size is {n_labels}.')
        data = dict(data)  # Disposing unwanted defaultdict side-effects
        return data, n_labels, frame_to_data_index_lookup

    def __len__(self) -> int:
        return self._n_labels



"""
Dataset factory method.
Use `DATASET_CATALOG.register` to extend supported dataset formats.
"""
from motrack.datasets.base import BaseDataset
from motrack.datasets.catalog import DATASET_CATALOG

# noinspection PyUnresolvedReferences
from motrack.datasets.mot import MOTDataset  # pylint: disable=unused-import


def dataset_factory(dataset_type: str, path: str, params: dict, test: bool = False) -> BaseDataset:
    """
    Dataset factory.

    Args:
        dataset_type: Dataset name (type)
        path: Path is mandatory parameter
        params: Dataset creation parameters
        test: If True then ground truth annotations are not required

    Returns:
        Created dataset object
    """
    if 'path' in params:
        raise ValueError('Key "path" must not be used in the params!')
    params['path'] = path
    if 'test' in params:
        raise ValueError('Key "test" must not be used in the params!')
    params['test'] = test

    return DATASET_CATALOG[dataset_type](**params)



"""
Momentum based association.
"""
from typing import Optional, List, Tuple

import numpy as np

from motrack.library.cv.bbox import PredBBox, Point
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.algorithms.utils import filter_observations
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG
from motrack.tracker.tracklet import Tracklet


class MomentumAssociation(AssociationAlgorithm):
    """
    Standard negative IoU association with gating.
    """
    def __init__(
        self,
        momentum: int = 1,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            momentum: Number of latest boxes (history) to consider for momentum association
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment)

        self._momentum = momentum

    def form_cost_matrix(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        _ = object_features  # Unused

        n_tracklets, n_detections = len(tracklet_estimations), len(detections)
        cost_matrix = np.zeros(shape=(n_tracklets, n_detections), dtype=np.float32)
        for t_i in range(n_tracklets):
            tracklet = tracklets[t_i]
            tracklet_bbox = tracklet_estimations[t_i]

            for d_i in range(n_detections):
                det_bbox = detections[d_i]

                cost_matrix[t_i][d_i] = self._calc_momentum(tracklet, tracklet_bbox, det_bbox, self._momentum)

        return cost_matrix

    @staticmethod
    def _calc_direction(point: Point, reference_point: Point) -> Tuple[float, float]:
        """
        Calculates unit direction given the point and the reference point.

        Direction: reference_point -> point

        Args:
            point: Point
            reference_point: Reference point

        Returns:
            Unit vector direction
        """
        direction_yx = point.y - reference_point.y, point.x - reference_point.x
        norm = np.sqrt(direction_yx[0] ** 2 + direction_yx[1] ** 2) + 1e-6
        direction_yx = (direction_yx[0] / norm, direction_yx[1] / norm)
        return direction_yx

    @staticmethod
    def momentum_score(
        tracklet_point: Point,
        tracklet_reference_point: Point,
        det_point: Point,
        det_reference_point: Point,
        conf: float
    ) -> float:
        """
        Calculates momentum score based on the tracklet direction and the candidate detection direction.

        Args:
            tracklet_point: Tracklet last point
            tracklet_reference_point: Tracklet previous point
            det_point: Detection point
            det_reference_point: Detection previous point
            conf: Detection confidence

        Returns:
            Momentum (angle) score
        """
        theta_inertia_y, theta_inertia_x = MomentumAssociation._calc_direction(tracklet_point, tracklet_reference_point)
        theta_intention_y, theta_intention_x = MomentumAssociation._calc_direction(det_point, det_reference_point)

        # scalar product: <x, y>
        # cos(theta) = <x, y> / (|x| * |y|)
        # |x| = |y| = 1
        # => theta = arcos(<x, y>)
        diff_angle_cos = theta_inertia_y * theta_intention_y + theta_inertia_x * theta_intention_x
        diff_angle_cos = np.clip(diff_angle_cos, a_min=-1, a_max=1)
        diff_angle = np.arccos(diff_angle_cos)
        diff_angle = (np.pi / 2.0 - np.abs(diff_angle)) / np.pi  # 1 in case of perfect match, 0 in case of bad match
        return - conf * diff_angle

    def _calc_momentum(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox, momentum: int) -> float:
        """
        Calculates matching score (cost) between tracklet and a detection bounding box.

        Args:
            tracklet: Tracklet info
            tracklet_bbox: Tracklet bounding box
            det_bbox: Detection bounding box
            momentum: Momentum

        Returns:
            Momentum score
        """
        raise NotImplementedError('Score calculation is not implemented!')

@ASSOCIATION_CATALOG.register('ocm')
class OCM(MomentumAssociation):
    """
    Observation Centric Momentum. Reference: https://arxiv.org/pdf/2203.14360.pdf
    """
    def __init__(
        self,
        delta_t: int = 3,
        fast_linear_assignment: bool = False,
    ):
        """
        Args:
            delta_t: Time step between tracker bounding boxes to estimate the speed vector
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment, momentum=1)
        self._delta_t = delta_t

    def _calc_momentum(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox, momentum: int) -> float:
        observation_history = filter_observations(tracklet.history)
        current_frame_index = tracklet.frame_index + tracklet.lost_time + 1
        latest_observation_history = [(index, bbox) for index, bbox in observation_history if index >= current_frame_index - self._delta_t - 1]

        if len(latest_observation_history) < 2:
            if len(observation_history) < 2:
                return 0.0

            _, last_bbox = observation_history[-1]
            _, prev_bbox = observation_history[-2]
            _, after_prev_bbox = observation_history[-1]
        else:
            _, last_bbox = latest_observation_history[-1]
            _, prev_bbox = latest_observation_history[0]
            _, after_prev_bbox = latest_observation_history[1]

        return self.momentum_score(last_bbox.center, prev_bbox.center, det_bbox.center, after_prev_bbox.center, det_bbox.conf)


@ASSOCIATION_CATALOG.register('robust-ocm')
class RobustOCM(MomentumAssociation):
    """
    Robust OCM. Reference: https://arxiv.org/pdf/2308.00783.pdf
    """
    def __init__(
        self,
        momentum: int = 3,
        fast_linear_assignment: bool = False,
    ):
        """
        Args:
            momentum: Number of latest boxes (history) to consider for momentum association
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment, momentum=momentum)

    @staticmethod
    def _get_corner_points(bbox: PredBBox) -> List[Point]:
        """
        Gets all bounding box corner points.

        Args:
            bbox: Bounding box

        Returns:
            Bounding box 4 corner points
        """
        return [
            Point(bbox.upper_left.x, bbox.upper_left.y),  # Upper Left
            Point(bbox.bottom_right.x, bbox.upper_left.y),  # Upper right
            Point(bbox.bottom_right.x, bbox.bottom_right.y),  # Bottom right
            Point(bbox.upper_left.x, bbox.bottom_right.y)  # Bottom Left
        ]

    def _calc_momentum(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox, momentum: int) -> float:
        observation_history = filter_observations(tracklet.history)
        current_frame_index = tracklet.frame_index + tracklet.lost_time + 1
        latest_observation_history = [(index, bbox) for index, bbox in observation_history if index >= current_frame_index - self._momentum - 1]

        if len(latest_observation_history) < 2:
            if len(observation_history) < 2:
                return 0.0

            _, curr_bbox = observation_history[-1]
            _, prev_bbox = observation_history[-2]
            _, after_prev_bbox = observation_history[-1]

            momentum_score: float = 0.0
            for curr_point, prev_point, det_point, after_prev_point in zip(self._get_corner_points(curr_bbox), self._get_corner_points(prev_bbox),
                                                         self._get_corner_points(det_bbox), self._get_corner_points(after_prev_bbox)):
                momentum_score += self.momentum_score(curr_point, prev_point, det_point, after_prev_point, det_bbox.conf)

            return momentum_score / 4

        momentum_score: float = 0.0
        for i in range(1, self._momentum + 1):
            if i >= len(latest_observation_history):
                continue

            _, curr_bbox = latest_observation_history[-1]
            _, prev_bbox = latest_observation_history[-i - 1]
            _, after_prev_bbox = latest_observation_history[-i]

            for curr_point, prev_point, det_point, after_prev_point in zip(self._get_corner_points(curr_bbox), self._get_corner_points(prev_bbox),
                                                         self._get_corner_points(det_bbox), self._get_corner_points(after_prev_bbox)):
                momentum_score += self.momentum_score(curr_point, prev_point, det_point, after_prev_point, det_bbox.conf)

        return momentum_score / 4


@ASSOCIATION_CATALOG.register('ecm')
class ECM(OCM):
    """
    Estimation Centric Momentum. Like OCM but uses estimation instead of the last observation.
    """
    def _calc_momentum(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox, momentum: int) -> float:
        if len(tracklet.history) < self._delta_t:
            prev_bbox = tracklet.bbox
        else:
            _, prev_bbox = tracklet.history[-self._delta_t]
        return self.momentum_score(tracklet_bbox.center, prev_bbox.center, det_bbox.center, prev_bbox.center, det_bbox.conf)


@ASSOCIATION_CATALOG.register('robust-ecm')
class RobustECM(RobustOCM):
    """
    Like Robust OCM but uses estimation instead of the last observation.
    """
    def _calc_momentum(self, tracklet: Tracklet, tracklet_bbox: PredBBox, det_bbox: PredBBox, momentum: int) -> float:
        if len(tracklet.history) < momentum:
            prev_bbox = tracklet.bbox

            momentum_score: float = 0.0
            for tracklet_point, prev_point, det_point in zip(self._get_corner_points(tracklet_bbox), self._get_corner_points(prev_bbox),
                                                             self._get_corner_points(det_bbox)):

                momentum_score += self.momentum_score(tracklet_point, prev_point, det_point, prev_point, det_bbox.conf)

            return momentum_score / 4

        momentum_score: float = 0.0
        for i in range(1, self._momentum + 1):
            _, prev_bbox = tracklet.history[-i]

            for tracklet_point, prev_point, det_point in zip(self._get_corner_points(tracklet_bbox), self._get_corner_points(prev_bbox),
                                             self._get_corner_points(det_bbox)):
                momentum_score += self.momentum_score(tracklet_point, prev_point, det_point, prev_point, det_bbox.conf)

        return momentum_score / 4



"""
Tracker inference postprocess (offline). Output can directly be evaluated using the TrackEval repo.
"""
import logging
import os

import hydra
from omegaconf import DictConfig

from motrack.common.project import CONFIGS_PATH
from motrack.config_parser import GlobalConfig
from motrack.datasets import dataset_factory
from motrack.tools.postprocess import run_tracker_postprocess
from motrack.utils import pipeline

logger = logging.getLogger('TrackerVizualization')


@pipeline.task('postprocess')
def run_postprocess(cfg: GlobalConfig) -> None:
    assert os.path.exists(cfg.experiment_path), f'Path "{cfg.experiment_path}" does not exist!'
    logger.info(f'Postprocessing tracker inference on path "{cfg.experiment_path}".')

    tracker_all_output = os.path.join(cfg.experiment_path, 'all')
    tracker_active_output = os.path.join(cfg.experiment_path, 'active')
    tracker_postprocess_output = os.path.join(cfg.experiment_path, 'postprocess')

    dataset = dataset_factory(
        dataset_type=cfg.dataset.type,
        path=cfg.dataset.fullpath,
        params=cfg.dataset.params,
        test=cfg.eval.split == 'test'
    )

    run_tracker_postprocess(
        dataset=dataset,
        tracker_active_output=tracker_active_output,
        tracker_all_output=tracker_all_output,
        tracker_postprocess_output=tracker_postprocess_output,
        postprocess_cfg=cfg.postprocess,
        scene_pattern=cfg.dataset_filter.scene_pattern
    )


@hydra.main(config_path=CONFIGS_PATH, config_name='movesort', version_base='1.1')
def main(cfg: DictConfig):
    # noinspection PyTypeChecker
    run_postprocess(cfg)


if __name__ == '__main__':
    main()



"""
NoMotion filter (just uses the last detection).
"""
from typing import Tuple

import numpy as np

from motrack.filter.algorithms.base import StateModelFilter, State
from motrack.filter.catalog import FILTER_CATALOG
from motrack.library.numpy_utils.bbox import affine_transform


@FILTER_CATALOG.register('no-motion')
class NoMotionFilter(StateModelFilter):
    """
    Baseline filter trusts fully detector and is completely certain in detector accuracy.
    """
    def __init__(self, det_uncertainty = 1e-6):
        self._det_uncertainty = det_uncertainty

    def initiate(self, measurement: np.ndarray) -> State:
        return measurement

    def predict(self, state: State) -> State:
        measurement = state
        return measurement

    def multistep_predict(self, state: State, n_steps: int) -> State:
        measurement = state
        return np.stack([measurement for _ in range(n_steps)])

    def update(self, state: State, measurement: np.ndarray) -> State:
        _ = state  # ignore previous state
        return measurement

    def missing(self, state: State) -> State:
        return state

    def project(self, state: State) -> Tuple[np.ndarray, np.ndarray]:
        measurement = state
        det_mat = np.ones_like(measurement) * self._det_uncertainty
        return state, det_mat

    def singlestep_to_multistep_state(self, state: State) -> State:
        return state.unsqueeze(0)

    def affine_transform(self, state: State, warp: np.ndarray) -> State:
        measurement = state
        return affine_transform(measurement, warp)



"""
Implementation of SparseTrack.
Reference: https://arxiv.org/abs/2306.05238
"""
from typing import Optional, Dict, Any

from motrack.tracker.trackers.algorithms.sort import SortTracker
from motrack.tracker.trackers.catalog import TRACKER_CATALOG


@TRACKER_CATALOG.register('fast')
class FastTracker(SortTracker):
    """
    FastTracker = Just IoU without any filters
    """
    def __init__(
        self,
        filter_name: str = 'no-motion',
        filter_params: Optional[dict] = None,
        cmc_name: Optional[str] = None,
        cmc_params: Optional[dict] = None,
        reid_name: Optional[str] = None,
        reid_params: Optional[str] = None,
        matcher_algorithm: str = 'default',
        matcher_params: Optional[Dict[str, Any]] = None,
        remember_threshold: int = 1,
        initialization_threshold: int = 3,
        new_tracklet_detection_threshold: Optional[float] = None,
        use_observation_if_lost: bool = False
    ):
        """
        Args:
            filter_name: Filter name
            filter_params: Filter params
            cmc_name: CMC name
            cmc_params: CMC params
            reid_name: ReID name
            reid_params: ReID params

            matcher_algorithm: Choose matching algorithm (e.g. Hungarian IOU)
            matcher_params: Matching algorithm parameters

            remember_threshold: How long does the tracklet without any detection matching
                If tracklet isn't matched for more than `remember_threshold` frames then
                it is deleted.
            initialization_threshold: Number of frames until tracklet becomes active
            new_tracklet_detection_threshold: Threshold to accept new tracklet
            use_observation_if_lost: When re-finding tracklet, use observation instead of estimation
        """
        if filter_params is None:
            filter_params = {}

        if matcher_algorithm == 'default':
            matcher_algorithm = 'iou'
            matcher_params = {
                'match_threshold': 0.2,
                'fast_linear_assignment': True
            }

        super().__init__(
            filter_name=filter_name,
            filter_params=filter_params,
            matcher_algorithm=matcher_algorithm,
            matcher_params=matcher_params,
            cmc_name=cmc_name,
            cmc_params=cmc_params,
            reid_name=reid_name,
            reid_params=reid_params,
            remember_threshold=remember_threshold,
            initialization_threshold=initialization_threshold,
            new_tracklet_detection_threshold=new_tracklet_detection_threshold,
            use_observation_if_lost=use_observation_if_lost
        )



"""
Object detection interface.
"""
from motrack.object_detection.factory import object_detection_inference_factory
from motrack.object_detection.algorithms.base import ObjectDetectionInference
from motrack.object_detection.algorithms.yolox import YOLOXInference
from motrack.object_detection.algorithms.yolov8 import YOLOv8Inference
from motrack.object_detection.manager import DetectionManager



"""
Implementation of ConfidenceFilter.
"""
import numpy as np
import scipy.linalg

# pylint: disable=pointless-string-statement
"""
Table for the 0.95 quantile of the chi-square distribution with N degrees of
freedom (contains values for N=1, ..., 9). Taken from MATLAB/Octave's chi2inv
function and used as Mahalanobis gating threshold.
"""
chi2inv95 = {
    1: 3.8415,
    2: 5.9915,
    3: 7.8147,
    4: 9.4877,
    5: 11.070,
    6: 12.592,
    7: 14.067,
    8: 15.507,
    9: 16.919
}


class ConfidenceKalmanFilter(object):
    """
    A simple Kalman filter for tracking detection confidence.

    The 2-dimensional state space

        c, vc
    """

    def __init__(
        self,
        initial_P_conf: float = 10.0,
        Q_conf: float = 1.0,
        Q_conf_velocity: float = 1e-3,
        R_conf: float = 100.0
    ):
        ndim, dt = 1, 1.

        # Create Kalman filter model matrices.
        self._motion_mat = np.eye(2 * ndim, 2 * ndim)
        for i in range(ndim):
            self._motion_mat[i, ndim + i] = dt
        self._update_mat = np.eye(ndim, 2 * ndim)

        # Hyperparameters
        self._initial_P_conf = initial_P_conf
        self._Q_conf = Q_conf
        self._Q_conf_velocity = Q_conf_velocity
        self._R_conf = R_conf

    def initiate(self, measurement):
        mean_pos = measurement
        mean_vel = np.zeros_like(mean_pos)
        mean = np.r_[mean_pos, mean_vel]

        std = [self._initial_P_conf, 1000 * self._initial_P_conf]
        covariance = np.diag(np.square(std))
        return mean, covariance

    def predict(self, mean, covariance):
        std_pos = [self._Q_conf]
        std_vel = [self._Q_conf_velocity]
        motion_cov = np.diag(np.square(np.r_[std_pos, std_vel]))

        mean = np.dot(mean, self._motion_mat.T)
        covariance = np.linalg.multi_dot((
            self._motion_mat, covariance, self._motion_mat.T)) + motion_cov

        return mean, covariance

    def project(self, mean, covariance):
        conf = mean[0]
        std = [self._R_conf * (1 - conf)]
        innovation_cov = np.diag(np.square(std))

        mean = np.dot(self._update_mat, mean)
        covariance = np.linalg.multi_dot((
            self._update_mat, covariance, self._update_mat.T))
        return mean, covariance + innovation_cov

    def update(self, mean, covariance, measurement):
        projected_mean, projected_cov = self.project(mean, covariance)

        # noinspection PyUnresolvedReferences
        chol_factor, lower = scipy.linalg.cho_factor(
            projected_cov, lower=True, check_finite=False)
        # noinspection PyUnresolvedReferences
        kalman_gain = scipy.linalg.cho_solve(
            (chol_factor, lower), np.dot(covariance, self._update_mat.T).T,
            check_finite=False).T
        innovation = measurement - projected_mean

        new_mean = mean + np.dot(innovation, kalman_gain.T)
        new_covariance = covariance - np.linalg.multi_dot((
            kalman_gain, projected_cov, kalman_gain.T))
        return new_mean, new_covariance


def run_test() -> None:
    measurements = np.array([0.1, 0.2, 0.3, 0.4, 0.1, 0.1, 0.1, 0.2, 0.8, 0.8, 0.9, 0.7, 0.6], dtype=np.float32).reshape(-1, 1)
    kf = ConfidenceKalmanFilter()
    mean, std = kf.initiate(measurements[0])

    for m in measurements[1:]:
        mean, std = kf.predict(mean, std)
        prior, _ = kf.project(mean, std)
        mean, std = kf.update(mean, std, m)
        posterior, _ = kf.project(mean, std)
        print(m, prior, posterior)


if __name__ == '__main__':
    run_test()



"""
MP4Reader.
"""
import os
from typing import Tuple, Union, Generator

import cv2
import numpy as np


class MP4Reader:
    """
    RAII mp4 reader
    """
    def __init__(self, path: str):
        """
        Args:
            path: Path to mp4 file

        Raises:
            FileNotFoundError: if video or sync file (meta file) does not exist
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f'Mp4 file {path} does not exist')

        self._path = path

        # State
        self._mp4_reader = None

    def next(self) -> Union[np.ndarray, None]:
        if not self._mp4_reader.isOpened():
            raise StopIteration
        ret, frame = self._mp4_reader.read()
        if not ret:
            return None

        return frame

    def open(self) -> 'MP4Reader':
        # noinspection PyUnresolvedReferences
        self._mp4_reader = cv2.VideoCapture(self._path)
        return self

    def close(self) -> None:
        if self._mp4_reader is not None:
            self._mp4_reader.release()

    def __enter__(self) -> 'MP4Reader':
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def iter(self) -> Generator[Tuple[int, np.ndarray], None, None]:
        """
        Returns:
            Frames generator (reading from video)
        """
        step_index = 0

        self.open()

        while True:
            try:
                frame = self.next()
                if frame is None:
                    break

                yield (step_index + 1), frame
            except StopIteration:
                break

            step_index += 1

        self.close()



"""
A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

# Get the version
main_ns = {}
ver_file = (here / 'motrack' / 'version.py').read_text()
exec(ver_file, main_ns)

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='motrack',  # Required
    version=main_ns['__version__'],  # Required
    description='Tracking-by-detection (MOT) package',  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional
    url='https://github.com/Robotmurlock/Motrack',  # Optional
    author='Momir Adzemovic',  # Optional
    author_email='momir.adzemovic@gmail.com',  # Optional
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='tracking-by-detection, multi-object-tracking',  # Optional
    packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests', 'scripts']),  # Required
    python_requires='>=3.8, <4',
    install_requires=[
        'hydra-core',
        'matplotlib',
        'numpy',
        'omegaconf',
        'opencv_python',
        'pandas',
        'PyYAML',
        'scipy',
        'tqdm',
        'torch',
        'motrack-motion'
    ],  # Optional
    extras_require={  # Optional
        'yolov8': ['ultralytics'],
        'reid': ['onnxruntime']
    }
)



"""
Implementation of SparseTrack.
Reference: https://arxiv.org/abs/2306.05238
"""
from typing import Optional, Dict, Any

from motrack.tracker.trackers.algorithms.byte import ByteTracker
from motrack.tracker.trackers.catalog import TRACKER_CATALOG


@TRACKER_CATALOG.register('sparse')
class SparseTracker(ByteTracker):
    """
    SparseTrack = ByteTrack + DCM
    """
    def __init__(
        self,
        filter_name: str = 'bot-sort',
        filter_params: Optional[dict] = None,
        cmc_name: Optional[str] = None,
        cmc_params: Optional[dict] = None,
        reid_name: Optional[str] = None,
        reid_params: Optional[str] = None,
        high_matcher_algorithm: str = 'default',
        high_matcher_params: Optional[Dict[str, Any]] = None,
        low_matcher_algorithm: str = 'default',
        low_matcher_params: Optional[Dict[str, Any]] = None,
        new_matcher_algorithm: str = 'default',
        new_matcher_params: Optional[Dict[str, Any]] = None,
        detection_threshold: float = 0.6,
        remember_threshold: int = 1,
        initialization_threshold: int = 3,
        new_tracklet_detection_threshold: Optional[float] = None,
        duplicate_iou_threshold: float = 0.85,
        use_observation_if_lost: bool = False
    ):
        if filter_params is None:
            filter_params = {}

        if high_matcher_algorithm == 'default':
            assert high_matcher_params is None
            high_matcher_algorithm = 'dcm'
            high_matcher_params = {
                'levels': 12,  # Default for DanceTrack
                'match_threshold': 0.2,
                'fuse_score': True
            }

        if low_matcher_algorithm == 'default':
            assert low_matcher_params is None
            low_matcher_algorithm = 'dcm'
            low_matcher_params = {
                'levels': 12,  # Default for DanceTrack
                'match_threshold': 0.5
            }

        if new_matcher_algorithm == 'default':
            assert new_matcher_params is None
            new_matcher_algorithm = 'iou'
            new_matcher_params = {
                'match_threshold': 0.3,
                'fuse_score': True
            }

        super().__init__(
            filter_name=filter_name,
            filter_params=filter_params,
            cmc_name=cmc_name,
            cmc_params=cmc_params,
            reid_name=reid_name,
            reid_params=reid_params,
            high_matcher_algorithm=high_matcher_algorithm,
            high_matcher_params=high_matcher_params,
            low_matcher_algorithm=low_matcher_algorithm,
            low_matcher_params=low_matcher_params,
            new_matcher_algorithm=new_matcher_algorithm,
            new_matcher_params=new_matcher_params,
            detection_threshold=detection_threshold,
            remember_threshold=remember_threshold,
            initialization_threshold=initialization_threshold,
            new_tracklet_detection_threshold=new_tracklet_detection_threshold,
            duplicate_iou_threshold=duplicate_iou_threshold,
            use_observation_if_lost=use_observation_if_lost
        )



"""
Dynamic LookupTable (class mappings).
"""
from typing import Dict, Optional, Iterable, Tuple, List


class LookupTable:
    """Vocabulary - Label lookup table (token <-> index)."""

    def __init__(
            self,
            token_to_index: Optional[Dict[str, int]] = None,
            unknown_token: str = '<unk>',
            add_unknown_token: bool = True
    ):
        """
        Args:
            token_to_index: Predefine token to index mappings.
            unknown_token: Unknown token value.
            add_unknown_token: Use unknown token.
        """
        self._token_to_index = token_to_index
        self._unknown_token = unknown_token
        self._add_unknown_token = add_unknown_token

        if self._token_to_index is None:
            self._token_to_index = {}

        if unknown_token not in self._token_to_index and add_unknown_token:
            self._token_to_index[unknown_token] = len(self._token_to_index)

        self._index_to_token = {v: k for k, v in self._token_to_index.items()}
        self._next_index = len(self._token_to_index)

    def add(self, token: str) -> int:
        """
        Adds token to the lookup table if it does not already exist.

        Args:
            token: Label (token)

        Returns:
            label (token) index
        """
        if token in self._token_to_index:
            return self._token_to_index[token]

        new_index = self._next_index
        self._next_index += 1
        self._token_to_index[token] = new_index
        self._index_to_token[new_index] = token
        return new_index

    def lookup(self, token: str) -> int:
        """
        Acquires token index if it exists in the table.
        In case the token does not exist in the lookup table:
            - and unknown token is used then unknown token index is returned;
            - otherwise KeyError is raised

        Raises:
            KeyError: Unknown token

        Returns:
            label (token) index
        """
        if token not in self._token_to_index and self._add_unknown_token:
            return self._token_to_index[self._unknown_token]

        return self._token_to_index[token]

    def inverse_lookup(self, index: int) -> str:
        """
        Inverse to `lookup`. Acquire token by index.

        Raises:
            KeyError: Unknown index

        Returns:
            label (token)
        """
        return self._index_to_token[index]

    def __iter__(self) -> Iterable[Tuple[str, int]]:
        for token, index in self._token_to_index.items():
            yield token, index

    @property
    def tokens(self) -> List[str]:
        """
        Returns:
            List of tokens in lookup table
        """
        return [token for token, _ in self]

    def __getitem__(self, token: str) -> int:
        return self.lookup(token)  # Alias for `lookup`

    def __len__(self):
        return self._next_index

    def serialize(self) -> dict:
        return {
            'token_to_index': self._token_to_index,
            'unknown_token': self._unknown_token,
            'add_unknown_token': self._add_unknown_token
        }

    @classmethod
    def deserialize(cls, raw: dict) -> 'LookupTable':
        return cls(**raw)


def run_test() -> None:
    lookup = LookupTable()
    lookup.add('dog')
    lookup.add('cat')

    assert lookup['dog'] == lookup.lookup('dog') == 1
    assert lookup['cat'] == lookup.lookup('cat') == 2
    assert lookup.inverse_lookup(1) == 'dog'
    assert lookup.inverse_lookup(2) == 'cat'
    assert lookup.tokens == ['<unk>', 'dog', 'cat']

    deserialized = LookupTable.deserialize(lookup.serialize())

    assert deserialized['dog'] == deserialized.lookup('dog') == 1
    assert deserialized['cat'] == deserialized.lookup('cat') == 2
    assert deserialized.inverse_lookup(1) == 'dog'
    assert deserialized.inverse_lookup(2) == 'cat'


if __name__ == '__main__':
    run_test()



"""
Tracker factory method.
Use `TRACKER_CATALOG.register` to extend supported tracker algorithms.
"""
from motrack.tracker.trackers.algorithms.base import Tracker
# noinspection PyUnresolvedReferences
from motrack.tracker.trackers.algorithms.byte import ByteTracker  # pylint: disable=unused-import
# noinspection PyUnresolvedReferences
from motrack.tracker.trackers.algorithms.sort import SortTracker  # pylint: disable=unused-import
# noinspection PyUnresolvedReferences
from motrack.tracker.trackers.algorithms.sparse import SparseTracker  # pylint: disable=unused-import
# noinspection PyUnresolvedReferences
from motrack.tracker.trackers.algorithms.fast import FastTracker  # pylint: disable=unused-import
from motrack.tracker.trackers.catalog import TRACKER_CATALOG


def tracker_factory(name: str, params: dict) -> Tracker:
    """
    Tracker factory

    Args:
        name: tracker name
        params: tracker params

    Returns:
        Tracker object
    """
    return TRACKER_CATALOG[name](**params)



"""
Support module for image drawing.
"""
import cv2
import numpy as np
from typing import List, Union

from motrack.library.cv import color_palette


def draw_text(
    image: np.ndarray,
    text: Union[str, List[str]],
    left: int,
    top: int,
    color: color_palette.ColorType,
    font_face: int = 1,
    font_scale: int = 1,
    thickness: int = 1,
    border_color: color_palette.ColorType = color_palette.BLACK,
    border_thickness: int = 2,
    vertical_label_step: int = 12,
) -> np.ndarray:
    """
    Draws text with border. With this heuristic it is easier to see image anywhere.

    Args:
        image: Image
        text: Text
        left: Text left position
        top: Text top position
        color: Text color
        font_face: Font face
        font_scale: Font scale
        thickness: Text thickness
        border_color: Text border color
        border_thickness: Text border thickness
        vertical_label_step: Space between two text labels
    Returns:
        Image with text
    """
    if isinstance(text, str):
        text = [text]

    n_labels = len(text)
    for i, label in enumerate(text):
        label_top = top - vertical_label_step * (n_labels - i)
        # noinspection PyUnresolvedReferences
        image = cv2.putText(image, label, (left, label_top), font_face, font_scale, border_color, border_thickness)
        # noinspection PyUnresolvedReferences
        image = cv2.putText(image, label, (left, label_top), font_face, font_scale, color, thickness)

    return image



"""
YOLOv8 inference support.
"""
from typing import List, Optional, Tuple, Union, Any

import numpy as np

from motrack.library.cv.bbox import LabelType
from motrack.object_detection.algorithms.base import ObjectDetectionInference
from motrack.object_detection.catalog import OBJECT_DETECTION_CATALOG
from motrack.utils.lookup import LookupTable


@OBJECT_DETECTION_CATALOG.register('yolov8')
class YOLOv8Inference(ObjectDetectionInference):
    """
    Object Detection - YOLOv8
    """
    def __init__(
        self,
        model_path: str,
        accelerator: str,
        verbose: bool = False,
        conf: float = 0.25,
        class_filter: Optional[List[LabelType]] = None,
        lookup: Optional[LookupTable] = None,
    ):
        super().__init__(lookup=lookup)
        try:
            import ultralytics
        except ImportError as e:
            raise ImportError('Please install ultralytics package with "pip3 install ultralytics" in order to use YOLOv8!') from e

        self._yolo = ultralytics.YOLO(model_path)
        self._yolo.to(accelerator)
        self._verbose = verbose
        self._conf = conf
        self._class_filter = class_filter
        if self._class_filter is None:
            self._class_filter = []

    def predict_raw(self, image: np.ndarray) -> Any:
        return self._yolo.predict(
            source=image,
            verbose=self._verbose,
            conf=self._conf,
            classes=self._class_filter
        )

    def postprocess(self, image: np.ndarray, raw: Any) -> Tuple[Union[np.ndarray, List[float]], List[LabelType], Union[np.ndarray, list]]:
        raw = raw[0]  # Remove batch
        h, w, _ = image.shape

        # Process bboxes
        bboxes = raw.boxes.xyxy.detach().cpu().numpy()
        bboxes[:, [0, 2]] /= w
        bboxes[:, [1, 3]] /= h

        # Process classes
        classes = raw.boxes.cls.detach().cpu().numpy()

        # Process confidences
        confidences = raw.boxes.conf.detach().cpu().view(-1).numpy()

        return bboxes, classes, confidences



"""
Standard DeepSORT association method based on IoU and appearance info.
"""
from typing import List
from typing import Optional

import numpy as np
from scipy.spatial.distance import cdist

from motrack.library.cv import PredBBox
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.algorithms.compose import ComposeAssociationAlgorithm
from motrack.tracker.matching.algorithms.iou import LabelGatingType, IoUAssociation
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG
from motrack.tracker.tracklet import TrackletCommonData, Tracklet


@ASSOCIATION_CATALOG.register('reid')
class ReIDAssociation(AssociationAlgorithm):
    """
    Standard ReID association method appearance.
    """
    def __init__(
        self,
        appearance_threshold: float = 0.0,
        appearance_metric: str = 'cosine',
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            appearance_threshold: Appearance metric minimum threshold
                - Disabled by default
            appearance_metric: Appearance metric (default: cosine)
            fast_linear_assignment: Use faster linear assignment
        """
        super().__init__(
            fast_linear_assignment=fast_linear_assignment,
        )

        self._appearance_threshold = appearance_threshold
        self._appearance_metric: str = appearance_metric

    def form_cost_matrix(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        assert tracklets is not None, 'Tracklets are mandatory for ReID based association!'
        assert object_features is not None, 'Detection features are also mandatory!'

        n_detections = object_features.shape[0]
        n_tracklets = len(tracklets)

        if n_tracklets == 0 or n_detections == 0:
            return np.zeros(shape=(n_tracklets, n_detections), dtype=np.float32)

        tracklet_features = self._get_features(tracklets)
        # noinspection PyTypeChecker
        cost_matrix = np.maximum(0.0, cdist(tracklet_features, object_features, metric=self._appearance_metric))
        cost_matrix[cost_matrix > 1 - self._appearance_threshold] = np.inf
        return cost_matrix

    @staticmethod
    def _get_features(tracklets: List[Tracklet]) -> np.ndarray:
        """
        Extracts appearance features from the tracklets.

        Args:
            tracklets: List of tracklets

        Returns:
            Tracklet appearance features
        """
        return np.stack([t.get(TrackletCommonData.APPEARANCE) for t in tracklets])


@ASSOCIATION_CATALOG.register('long-term-reid')
class LongTermReIdAssociation(ReIDAssociation):
    """
    Long-term based ReID association.
    """
    def __init__(
        self,
        appearance_threshold: float = 0.0,
        appearance_metric: str = 'cosine',
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            appearance_threshold: Appearance metric minimum threshold
                - Disabled by default
            appearance_metric: Appearance metric (default: cosine)
            fast_linear_assignment: Use faster linear assignment
        """
        super().__init__(
            fast_linear_assignment=fast_linear_assignment,
            appearance_threshold=appearance_threshold,
            appearance_metric=appearance_metric
        )

    @staticmethod
    def _get_features(tracklets: List[Tracklet]) -> np.ndarray:
        return np.stack([np.stack(list(t.get(TrackletCommonData.APPEARANCE_BUFFER))).mean(axis=0) for t in tracklets])


@ASSOCIATION_CATALOG.register('reid-iou')
class ReIDIoUAssociation(ComposeAssociationAlgorithm):
    """
    Standard DeepSORT association method based on IoU and appearance.
    """
    def __init__(
        self,
        appearance_weight: float = 0.5,
        appearance_threshold: float = 0.0,
        appearance_metric: str = 'cosine',
        match_threshold: float = 0.30,
        fuse_score: bool = False,
        label_gating: Optional[LabelGatingType] = None,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            appearance_weight: Appearance to IoU association weight
            appearance_threshold: Appearance metric minimum threshold
            appearance_metric: Appearance metric (default: cosine)
            match_threshold: IoU match threshold
            fuse_score: Fuse detection score with IoU match
            label_gating: Allow different labels to be matched
            fast_linear_assignment: Use faster linear assignment
        """
        assert 0 <= appearance_weight <= 1.0, '"appearance_weight" must be in interval [0, 1]!'

        matchers = [
            IoUAssociation(
                match_threshold=match_threshold,
                fuse_score=fuse_score,
                label_gating=label_gating
            ),
            ReIDAssociation(
                appearance_threshold=appearance_threshold,
                appearance_metric=appearance_metric
            )
        ]

        weights = [1 - appearance_weight, appearance_weight]

        super().__init__(
            matchers=matchers,
            weights=weights,
            fast_linear_assignment=fast_linear_assignment
        )



"""
Set of path constants relative to project source (independent of project location on system).
"""
import os


ROOT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
SRC_PATH = os.path.join(ROOT_PATH, 'motrack')
CONFIGS_PATH = os.path.join(ROOT_PATH, 'configs')
OUTPUTS_PATH = os.path.join(ROOT_PATH, 'outputs')
PLAYGROUND_PATH = os.path.join(ROOT_PATH, 'playground')

ASSETS_PATH = '/media/home/motrack-outputs/'
MASTER_PATH = '/media/home/'



"""
Implementation of Confidence association inspired by the Hybrid-SORT paper.
Reference: https://arxiv.org/pdf/2308.00783.pdf
"""
from typing import List, Optional

import numpy as np
from scipy.spatial.distance import cdist

from motrack.library.cv import PredBBox
from motrack.library.kalman_filter.conf_kf import ConfidenceKalmanFilter
from motrack.tracker.matching.algorithms.base import AssociationAlgorithm
from motrack.tracker.matching.catalog import ASSOCIATION_CATALOG
from motrack.tracker.tracklet import Tracklet, TrackletState
from motrack.tracker.matching.algorithms.utils import filter_observations


@ASSOCIATION_CATALOG.register('hybrid-conf')
class HybridConfidenceAssociation(AssociationAlgorithm):
    """
    Hybrid-SORT inspired confidence based association algorithm.
    """
    DATA_KF_KEY = 'kf-conf-est'

    def __init__(
        self,
        initial_P_conf: float = 10.0,
        Q_conf: float = 1.0,
        Q_conf_velocity: float = 1e-3,
        R_conf: float = 100.0,
        linear_prediction: bool = False,
        lower_bound: float = 0.1,
        upper_bound: float = 0.9,
        fast_linear_assignment: bool = False
    ):
        """
        Args:
            initial_P_conf: Initial confidence uncertainty (std)
            Q_conf: Initial process noise confidence "position"
            Q_conf_velocity: Initial process noise confidence "velocity"
            R_conf: Confidence noise uncertainty
            linear_prediction: Use linear prediction instead of KF
            lower_bound: Lower confidence bound
            upper_bound: Upper confidence bound
            fast_linear_assignment: Use greedy algorithm for linear assignment
                - This might be more efficient in case of large cost matrix
        """
        super().__init__(fast_linear_assignment=fast_linear_assignment)
        self._kf = ConfidenceKalmanFilter(
            initial_P_conf=initial_P_conf,
            Q_conf=Q_conf,
            Q_conf_velocity=Q_conf_velocity,
            R_conf=R_conf
        )
        self._linear_prediction = linear_prediction
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def _postprocess_conf(self, conf: float) -> float:
        return max(self._lower_bound, min(self._upper_bound, conf))

    def _predict_conf(self, tracklet: Tracklet) -> float:
        """
        Predicts confidence for next step.
        Uses linear (naive) extrapolation or KF.

        Args:
            tracklet: Tracklet

        Returns:
            Predicted confidence
        """
        if self._linear_prediction:
            # Perform linear prediction
            observation_history = filter_observations(tracklet.history)

            history_len = len(observation_history)
            if history_len < 2:
                return self._postprocess_conf(tracklet.bbox.conf)

            last_bbox_frame_index, last_bbox = observation_history[-1]
            last_bbox_conf = last_bbox.conf
            prev_bbox_frame_index, prev_bbox = observation_history[-2]
            prev_bbox_conf = prev_bbox.conf

            if last_bbox_frame_index - prev_bbox_frame_index > 1:
                return self._postprocess_conf(tracklet.bbox.conf)

            delta = last_bbox_conf - prev_bbox_conf
            conf = last_bbox_conf + delta

            return self._postprocess_conf(conf)

        # Perform KF prediction
        mean, std = tracklet.get(self.DATA_KF_KEY)
        mean, std = self._kf.predict(mean, std)
        tracklet.set(self.DATA_KF_KEY, (mean, std))

        conf = float(self._kf.project(mean, std)[0])
        return self._postprocess_conf(conf)

    def _update_conf(self, tracklets: List[Tracklet]) -> None:
        """
        Updates KF state with new confidence "measurements".

        Args:
            tracklets: List of tracklets
        """
        for tracklet in tracklets:
            if tracklet.state == TrackletState.LOST:
                continue

            conf_measurement = np.array([tracklet.bbox.conf], dtype=np.float32)

            state = tracklet.get(self.DATA_KF_KEY)
            if state is None:
                mean, std = self._kf.initiate(conf_measurement)
            else:
                mean, std = state
                mean, std = self._kf.update(mean, std, conf_measurement)

            tracklet.set(self.DATA_KF_KEY, (mean, std))

    def form_cost_matrix(
        self,
        tracklet_estimations: List[PredBBox],
        detections: List[PredBBox],
        object_features: Optional[np.ndarray] = None,
        tracklets: Optional[List[Tracklet]] = None
    ) -> np.ndarray:
        _, _ = tracklet_estimations, object_features  # Unused

        self._update_conf(tracklets)
        tracklet_conf_estimations = np.array([self._predict_conf(t) for t in tracklets], dtype=np.float32).reshape(-1, 1)
        detection_confs = np.array([bbox.conf for bbox in detections], dtype=np.float32).reshape(-1, 1)
        return np.maximum(0.0, cdist(tracklet_conf_estimations, detection_confs, metric='cityblock'))



"""
Tracker inference tool.
"""
import re
from typing import List

from tqdm import tqdm

from motrack.datasets import BaseDataset
from motrack.evaluation.io import TrackerInferenceWriter
from motrack.object_detection import DetectionManager
from motrack.tracker import Tracker
from motrack.tracker.tracklet import Tracklet, TrackletState


def run_tracker_inference(
    dataset: BaseDataset,
    tracker: Tracker,
    detection_manager: DetectionManager,
    tracker_active_output: str,
    tracker_all_output: str,
    clip: bool = True,
    scene_pattern: str = '(.*?)',
    load_image: bool = True
) -> None:
    """
    Performs inference on given dataset with a given tracker and detection manager.

    Args:
        dataset: Dataset to perform tracker inference on
        tracker: Tracker
        detection_manager: Detection manager
        tracker_active_output: Path where the active tracks are stored
        tracker_all_output: Path where the all tracks are stored
        clip: Clip bounding boxes coordinates to range [0, 1]
        scene_pattern: Filter dataset scenes.
        load_image: Load image for Object Detection or ReID model
            - Can be set to False if everything is already cached
    """
    scene_names = dataset.scenes
    scene_names = [scene_name for scene_name in scene_names if re.match(scene_pattern, scene_name)]
    for scene_name in tqdm(scene_names, desc='Simulating tracker', unit='scene'):
        tracker.reset_state()
        tracker.set_scene(scene_name)
        tracklets: List[Tracklet] = []

        scene_info = dataset.get_scene_info(scene_name)
        scene_length = scene_info.seqlength
        imheight = scene_info.imheight
        imwidth = scene_info.imwidth

        with TrackerInferenceWriter(tracker_active_output, scene_name, image_height=imheight, image_width=imwidth,
                                    clip=clip) as tracker_active_inf_writer, \
                TrackerInferenceWriter(tracker_all_output, scene_name, image_height=imheight, image_width=imwidth,
                                       clip=clip) as tracker_all_inf_writer:
            for index in tqdm(range(scene_length), desc=f'Simulating "{scene_name}"', unit='frame'):
                # Perform OD inference
                detection_bboxes = detection_manager.predict(scene_name, index)

                # Perform tracking step
                tracklets = tracker.track(
                    tracklets=tracklets,
                    detections=detection_bboxes,
                    frame_index=index + 1,  # Counts from 1 instead of 0
                    frame=dataset.load_scene_image_by_frame_index(scene_name, index) if load_image else None
                )
                active_tracklets = [t for t in tracklets if t.state == TrackletState.ACTIVE]

                # Save inference
                for tracklet in active_tracklets:
                    tracker_active_inf_writer.write(index, tracklet)

                for tracklet in tracklets:
                    tracker_all_inf_writer.write(index, tracklet)
